package com.ningdatech.pmapi.meeting.task;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.convert.Convert;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.conditions.update.LambdaUpdateWrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.ningdatech.basic.util.CollUtils;
import com.ningdatech.cache.model.cache.CacheKey;
import com.ningdatech.cache.repository.CachePlusOps;
import com.ningdatech.pmapi.common.util.SpringContextHolder;
import com.ningdatech.pmapi.meeting.builder.ExpertInviteBuilder;
import com.ningdatech.pmapi.meeting.entity.domain.ExpertInviteRule;
import com.ningdatech.pmapi.meeting.entity.domain.Meeting;
import com.ningdatech.pmapi.meeting.entity.domain.MeetingExpert;
import com.ningdatech.pmapi.meeting.entity.dto.AvoidInfoDTO;
import com.ningdatech.pmapi.meeting.entity.dto.ExpertChooseDTO;
import com.ningdatech.pmapi.meeting.entity.dto.RandomInviteRuleDTO;
import com.ningdatech.pmapi.meeting.entity.enumeration.ExpertAttendStatusEnum;
import com.ningdatech.pmapi.meeting.entity.enumeration.ExpertInviteTypeEnum;
import com.ningdatech.pmapi.meeting.helper.ExpertInviteHelper;
import com.ningdatech.pmapi.meeting.helper.YxtCallOrSmsHelper;
import com.ningdatech.pmapi.meeting.manage.ExpertInviteManage;
import com.ningdatech.pmapi.meeting.service.IExpertInviteAvoidRuleService;
import com.ningdatech.pmapi.meeting.service.IExpertInviteRuleService;
import com.ningdatech.pmapi.meeting.service.IMeetingExpertService;
import com.ningdatech.pmapi.meeting.service.IMeetingService;
import com.ningdatech.pmapi.user.util.LoginUserUtil;
import lombok.AllArgsConstructor;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.springframework.aop.framework.AopContext;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.scheduling.concurrent.ThreadPoolTaskScheduler;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import java.time.Duration;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Function;
import java.util.stream.Collectors;


/**
 * <p>
 * ExpertInviteTask
 * </p>
 *
 * @author WendyYang
 * @since 21:25 2022/8/15
 */
@Slf4j
@Component
@AllArgsConstructor
public class ExpertInviteTask {

    private final ExpertInviteHelper expertInviteHelper;
    private final RandomInviteProperties properties;

    private static final String MEETING_ID_INVITE_RANDOM = "MEETING_ID_INVITE_RANDOM";
    private static final CacheKey CACHE_KEY = new CacheKey(MEETING_ID_INVITE_RANDOM);

    private final CachePlusOps cachePlusOps;
    @Qualifier("expertInviteScheduler")
    @Resource(name = "expertInviteScheduler")
    private ThreadPoolTaskScheduler scheduler;
    private final IMeetingExpertService meetingExpertService;
    private final IExpertInviteRuleService inviteRuleService;
    private final IMeetingService meetingService;
    private final ExpertInviteManage expertInviteManage;
    private final IExpertInviteAvoidRuleService expertInviteAvoidRuleService;

    private final YxtCallOrSmsHelper yxtCallOrSmsHelper;

    public ExpertInviteTask currProxy() {
        return (ExpertInviteTask) AopContext.currentProxy();
    }

    /**
     * 用来存入线程执行情况, 方便于停止定时任务时使用
     */
    protected static final ConcurrentMap<Long, ScheduledFuture<?>> INVITE_MAP = new ConcurrentHashMap<>();

    @PostConstruct
    public void initTask() {
        if (!properties.getEnable()) {
            log.warn("随机邀请已关闭……");
            return;
        }
//        initInviteTaskByStart();
    }

    /**
     * 项目重启之后重新初始化邀请任务
     */
    private void initInviteTaskByStart() {
        Set<Object> meetingIds = cachePlusOps.sMembers(CACHE_KEY);
        if (meetingIds == null || meetingIds.isEmpty()) {
            return;
        }
        for (Object meetingId : meetingIds) {
            addInviteExpertTask(Convert.toLong(meetingId), true, properties.getInviteDelay());
        }
    }

    /**
     * 专家抽取校验
     *
     * @param meetingId 会议ID
     * @return boolean
     * @author WendyYang
     **/
    private boolean inviteCheck(Long meetingId) {
        List<ExpertInviteRule> expertInviteRules = inviteRuleService.listByMeetingId(meetingId);
        Function<ExpertInviteRule, ExpertInviteTypeEnum> groupKey = w -> ExpertInviteTypeEnum.getByCode(w.getInviteType());
        Map<ExpertInviteTypeEnum, List<ExpertInviteRule>> groupByType = CollUtils.group(expertInviteRules, groupKey);
        List<ExpertInviteRule> randomRules = groupByType.get(ExpertInviteTypeEnum.RANDOM);
        if (CollUtil.isEmpty(randomRules)) {
            return false;
        }
        Map<Long, ExpertInviteRule> ruleMap = CollUtils.listToMap(randomRules, ExpertInviteRule::getId);
        LambdaQueryWrapper<MeetingExpert> query = Wrappers.lambdaQuery(MeetingExpert.class)
                .in(MeetingExpert::getRuleId, ruleMap.keySet())
                .in(MeetingExpert::getStatus, ExpertAttendStatusEnum.AGREED.getCode());
        List<MeetingExpert> meetingExperts = meetingExpertService.list(query);
        int totalCount = CollUtils.sum(randomRules, ExpertInviteRule::getInviteCount);
        boolean needed = totalCount > meetingExperts.size();
        if (!needed) {
            cancelByMeetingId(meetingId);
        }
        return needed;
    }

    /**
     * 唤醒某个会议的抽取任务
     *
     * @param meetingId 会议ID
     * @author WendyYang
     **/
    public void addInviteExpertTask(Long meetingId) {
        boolean contains = INVITE_MAP.containsKey(meetingId);
        if (!contains) {
            addInviteExpertTask(meetingId, false, properties.getInviteDelay());
            log.info("重置会议的随机抽取状态：{}", meetingId);
            LambdaUpdateWrapper<Meeting> update = Wrappers.lambdaUpdate(Meeting.class);
            update.set(Meeting::getInviteStatus, false);
            update.set(Meeting::getUpdateBy, LoginUserUtil.getUserId());
            update.set(Meeting::getUpdateOn, LocalDateTime.now());
            update.eq(Meeting::getId, meetingId);
            meetingService.update(update);
        }
    }

    /**
     * 添加专家抽取校验任务
     *
     * @param meetingId      会议ID
     * @param checked        是否前置校验
     * @param delayedMinutes 延迟执行时间
     * @author WendyYang
     **/
    public void addInviteExpertTask(Long meetingId, boolean checked, int delayedMinutes) {
        if (checked && !inviteCheck(meetingId)) {
            return;
        }
        Instant startTime = LocalDateTime.now().plusMinutes(delayedMinutes).atZone(ZoneId.systemDefault()).toInstant();
        ScheduledFuture<?> future = scheduler.scheduleAtFixedRate(() -> {
            ExpertInviteTask bean = SpringContextHolder.getBean(ExpertInviteTask.class);
            try {
                bean.invite(meetingId);
            } catch (Exception e) {
                log.error("执行专家邀请任务异常：{}", meetingId, e);
            }
        }, startTime, Duration.ofMinutes(properties.getInviteFixedRate()));
        INVITE_MAP.putIfAbsent(meetingId, future);
        log.info("添加专家抽取后台任务：{}", meetingId);
    }

    public void addInviteExpertTaskByMeetingCreate(Long meetingId, int delayedMinutes) {
        Assert.isTrue(properties.getEnable(), "随机邀请已关闭");
        addInviteExpertTask(meetingId, false, delayedMinutes);
        cachePlusOps.sAdd(CACHE_KEY, meetingId);
    }

    @Transactional(rollbackFor = Exception.class)
    public void invite(Long meetingId) {
        log.info("开始进行专家后台抽取：{}", meetingId);
        Meeting meeting = meetingService.getById(meetingId);
        if (meeting.getStartTime().isBefore(LocalDateTime.now())) {
            // 会议开始结束随机抽取
            cancelByMeetingId(meetingId);
            log.info("会议已开始停止抽取：{}", meeting);
            return;
        }
        // 随机邀请规则
        Map<Long, RandomInviteRuleDTO> ruleMap = inviteRuleService.randomRuleByMeetingId(meetingId);
        // 回避规则
        AvoidInfoDTO avoidInfoDto = expertInviteAvoidRuleService.getAvoidInfoDto(meetingId);
        // 还需要抽取的规则数量
        AtomicInteger notIgnoreCnt = new AtomicInteger(ruleMap.size());
        AtomicInteger notSupportCnt = new AtomicInteger(0);
        ruleMap.forEach((ruleId, value) -> {
            List<Long> singletonList = Collections.singletonList(meetingId);
            List<MeetingExpert> tempExperts = meetingExpertService.listExpertLastByMeetingIds(singletonList);
            Map<Long, MeetingExpert> expertMap = CollUtils.listToMap(tempExperts, MeetingExpert::getExpertId);
            Map<Long, MeetingExpert> replacedMap = expertMap.values().stream().filter(w -> w.getPreId() > 0)
                    .collect(Collectors.toMap(MeetingExpert::getPreId, w -> w));
            Map<Long, ExpertCntBO> countMap = countByAgree(expertMap, replacedMap);
            // 已确认参加、通话中数量
            ExpertCntBO cnt = countMap.getOrDefault(ruleId, ExpertCntBO.zeroInit());
            int tempCurrent = cnt.getAgreeCnt() + cnt.getNoticeCnt();
            if (tempCurrent == value.getCount()) {
                if (cnt.getAgreeCnt().equals(value.getCount())) {
                    notIgnoreCnt.decrementAndGet();
                }
                return;
            }
            int currInviteCnt = value.getCount() - tempCurrent;
            ExpertChooseDTO expertChoose = expertInviteManage.expertReplaceByRandomRule(avoidInfoDto, value,
                    expertMap.values(), currInviteCnt, meeting.getStartTime(), meeting.getEndTime(), null);

            if (expertChoose.getTotal() > 0) {
                List<MeetingExpert> expertMeetings = CollUtils.convert(expertChoose.getExperts(), w -> {
                    MeetingExpert expert = ExpertInviteBuilder.getExpertByRandom(meetingId, w, ruleId);
                    expert.setPreStatus(ExpertAttendStatusEnum.NOTICING.getCode());
                    expert.setStatus(ExpertAttendStatusEnum.NOTICING.getCode());
                    return expert;
                });
                yxtCallOrSmsHelper.callByMeetingExperts(meeting, expertMeetings);
                log.info("会议：{} 后台抽取专家：{}名", meetingId, expertMeetings.size());
                meetingExpertService.saveBatch(expertMeetings);
            } else {
                // 抽取人数不够
                notSupportCnt.incrementAndGet();
            }
        });
        if (notIgnoreCnt.get() == 0 || notIgnoreCnt.get() == notSupportCnt.get()) {
            if (notSupportCnt.get() > 0) {
                // TODO
                /*UserInfo inviterBasic = userInfoService.getById(meeting.getCreateBy());
                UserInfo inviter = userInfoService.getById(inviterBasic.getId());
                SendSmsContext context = YxtSmsContextBuilder.smsByRandomInviteStop(inviter.getNickname(), meeting.getName(), inviterBasic.getPhoneNo());
                yxtCallOrSmsHelper.sendSms(context);*/
            }
            log.info("停止会议随机邀请：{} 未完成抽取规则数量 {} 无可抽取专家规则数量 {}", meetingId, notIgnoreCnt, notSupportCnt);
            currProxy().cancelByMeetingId(meetingId);
        }
    }

    @Transactional(rollbackFor = Exception.class)
    public void cancelByMeetingId(Long meetingId) {
        log.info("终止专家抽取：{}", meetingId);
        meetingService.stopRandomInvite(meetingId);
        cachePlusOps.sRem(CACHE_KEY, meetingId);
        ScheduledFuture<?> future = INVITE_MAP.get(meetingId);
        if (future != null) {
            INVITE_MAP.remove(meetingId);
            if (!future.isCancelled()) {
                future.cancel(true);
            }
        }
    }

    //==================================================================================================================

    private Map<Long, ExpertCntBO> countByAgree(Map<Long, MeetingExpert> expertMap, Map<Long, MeetingExpert> replacedMap) {
        return expertMap.entrySet().stream()
                .collect(Collectors.groupingBy(w -> w.getValue().getRuleId(),
                        Collectors.collectingAndThen(Collectors.mapping(Map.Entry::getValue, Collectors.toList()), w -> {
                            ExpertCntBO cnt = ExpertCntBO.zeroInit();
                            for (MeetingExpert expert : w) {
                                if (ExpertAttendStatusEnum.AGREED.eq(expert.getStatus())) {
                                    cnt.incrAgreeCnt();
                                } else if (ExpertAttendStatusEnum.NOTICING.eq(expert.getStatus())) {
                                    cnt.incrNoticeCnt();
                                } else if (ExpertAttendStatusEnum.REPLACED.eq(expert.getStatus())) {
                                    MeetingExpert replacedExpert = replacedMap.get(expert.getId());
                                    if (replacedExpert != null && ExpertAttendStatusEnum.AGREED.eq(replacedExpert.getStatus())) {
                                        cnt.incrAgreeCnt();
                                    }
                                }
                            }
                            return cnt;
                        })));
    }

    @Data
    public static class ExpertCntBO {

        private Integer agreeCnt;

        private Integer noticeCnt;

        public ExpertCntBO(Integer agreeCnt, Integer noticeCnt) {
            this.agreeCnt = agreeCnt;
            this.noticeCnt = noticeCnt;
        }

        public void incrAgreeCnt() {
            agreeCnt++;
        }

        public void incrNoticeCnt() {
            noticeCnt++;
        }

        public static ExpertCntBO zeroInit() {
            return new ExpertCntBO(0, 0);
        }

    }

}
